import requests
import os
import datetime


DEFAULT_PARAMS = {"format": "json", "formatversion": "2"}

CHUNKSIZE = 1024 * 1024 * 4


class Wiki:
    def __init__(self, domain, username=None, password=None):
        self.endpoint = f"https://{domain}/w/api.php"
        self.domain = domain
        self.client = requests.Session()
        self.username = username

        self.csrf_token = None

        if username and password:
            self.login(username, password)

    def __make_params(self, action, pl={}):
        global DEFAULT_PARAMS
        d = DEFAULT_PARAMS.copy()

        d["action"] = action
        d.update(pl)
        return d

    def acceptable_file_extensions(self):
        """Gets acceptable file extensions which may be uploaded to this Wiki.  Returns a list with acceptable extensions"""

        ColorLog.info("Fetching a list of acceptable file upload extensions", self)

        response = self.client.get(self.endpoint, params=self.__make_params("query", {"meta": "siteinfo", "siprop": "fileextensions"}))
        return {jo["ext"] for jo in response.json()["query"]['fileextensions']}

    def login(self, username, password):
        """Logs a user in.

        :param username: the username to use
        :param password: The password to use
        """
        ColorLog.info("Try login for " + username, self)

        response = self.client.post(self.endpoint, params=self.__make_params("login"), data={
                                    "lgname": username, "lgpassword": password, "lgtoken": self.get_tokens()["logintoken"]})

        self.username = response.json()["login"]["lgusername"]
        self.csrf_token = self.get_tokens()['csrftoken']

        return True

    def get_tokens(self):
        """Gets CSRF and Login tokens.  Returns a dict with keys csrftoken and logintoken"""
        return self.client.get(self.endpoint, params=self.__make_params("query", {"meta": "tokens", "type": "csrf|login"})).json()['query']['tokens']

    def upload(self, path, title, desc, summary):
        """Uploads a file.  Returns True on success

        :param path: the local path to the file to upload
        :param title: The title to upload the file to, excluding "File:" namespace
        :param desc: The text of the description page
        :param summary: The edit summary to use
        """

        fsize = os.path.getsize(path)
        total_chunks = fsize // CHUNKSIZE + 1

        ColorLog.info("Uploading " + path, self)

        data = DEFAULT_PARAMS.copy()
        data.update({"filename": title, "offset": '0', "ignorewarnings": "1", "filesize": str(fsize), "token": self.csrf_token, "stash": "1"})

        with open(path, 'rb') as f:
            buffer = f.read(CHUNKSIZE)
            chunk_count = 0

            err_count = 0
            while True:
                if err_count > 5:
                    ColorLog.error(f"Encountered {err_count} errors, aborting", self)
                    break

                ColorLog.fyi(f"Uploading chunk [{chunk_count + 1} of {total_chunks}] of '{path}'", self)

                response = self.client.post(self.endpoint, params={"action": "upload"}, data=data, files={
                                            'chunk': (os.path.basename(path), buffer, "multipart/form-data")}, timeout=420)
                if not response:
                    ColorLog.error("Did not get a response back from the server, retrying...", self)
                    err_count += 1
                    continue

                response = response.json()
                if "error" in response:
                    ColorLog.error(response['error']['info'], self)
                    err_count += 1
                    continue

                data['filekey'] = response["upload"]["filekey"]
                chunk_count += 1
                data['offset'] = str(CHUNKSIZE * chunk_count)

                buffer = f.read(CHUNKSIZE)
                if not buffer:
                    break

        if "filekey" not in data:
            return False

        ColorLog.info(f"Unstashing {data['filekey']} as {title}", self)
        pl = {"filename": title, "text": desc, "comment": summary, "filekey": data['filekey'], "ignorewarnings": "1", "token": self.csrf_token}
        pl.update(DEFAULT_PARAMS)
        response = self.client.post(self.endpoint, params={"action": "upload"}, data=pl, timeout=420).json()

        if 'error' in response:
            ColorLog.error(response['error']['info'], self)
            return False

        return response['upload']['result'] == "Success"


class ColorLog:

    @staticmethod
    def __log(color_code, level, msg, wiki=None):
        """Prints out a message to std out
            :param color_code: The ANSI color to print (1-9)
            :param level: The level string (e.g. WARN, INFO, ERROR)
            :param msg: the mesage to print
            :param wiki: the Wiki to generate a prefix with (optional)
        """
        prefix = f"[{wiki.username or '<Anonymous>'} @ {wiki.domain}]: " if wiki else ""
        print(f"{datetime.datetime.now():%b %d, %Y %I:%M:%S %p}\n{level}: \033[3{color_code}m{prefix}{msg}\033[0m")

    @staticmethod
    def warn(msg, wiki=None):
        """Prints out a warning message.
            :param msg: the mesage to print
            :param wiki: the Wiki to generate a prefix with (optional)
        """
        ColorLog.__log(3, "WARN", msg, wiki)

    @staticmethod
    def info(msg, wiki=None):
        """Prints out an info message.
            :param msg: the mesage to print
            :param wiki: the Wiki to generate a prefix with (optional)
        """
        ColorLog.__log(2, "INFO", msg, wiki)

    @staticmethod
    def error(msg, wiki=None):
        """Prints out an error message.
            :param msg: the mesage to print
            :param wiki: the Wiki to generate a prefix with (optional)
        """
        ColorLog.__log(1, "ERROR", msg, wiki)

    @staticmethod
    def fyi(msg, wiki=None):
        """Prints out an FYI message.
            :param msg: the mesage to print
            :param wiki: the Wiki to generate a prefix with (optional)
        """
        ColorLog.__log(6, "FYI", msg, wiki)



import argparse
import getpass
import re
from datetime import date, datetime
from pathlib import Path
from textwrap import dedent

from PIL import ExifTags, Image

from scu.wiki import Wiki, ColorLog
from scu.wgen import Wgen


def main():
    MTC_FILE = Path.home() / ".scu.px.txt"

    cli_parser = argparse.ArgumentParser(description="Simple Commons Uploader")
    cli_parser.add_argument('dirs', metavar='folders', type=str, nargs='*', help='folders with files to upload')
    cli_parser.add_argument('--user', type=str, help="username to use")
    cli_parser.add_argument('--pw', type=str, help="password to use")
    cli_parser.add_argument("-i", action='store_true', help="force interactive login")
    cli_parser.add_argument("--wgen", action='store_true', help="run wgen password manager")
    args = cli_parser.parse_args()

    if args.wgen:
        Wgen.setup(MTC_FILE, False)
        return

    wiki = Wiki("commons.wikimedia.org")
    if args.i:
        wiki.login(input("Please login to continue.\nUsername: "), getpass.getpass())
    elif args.user:
        if not args.pw:
            ColorLog.error("You didn't specify a password, --pw")
            return
        wiki.login(args.user, args.pw)
    elif MTC_FILE.is_file():
        pxd = Wgen.load_px(MTC_FILE)
        if not pxd:
            ColorLog.error(f"Please run with --wgen option or remove {MTC_FILE} from your home directory")
        wiki.login(*pxd.popitem())
    else:
        cli_parser.print_help()
        return

    if not args.dirs:
        ColorLog.warn("You didn't specify and directories to upload!  Goodbye.")
        return

    ext_list = {"." + e for e in wiki.acceptable_file_extensions()}

    for d in args.dirs:
        base_dir = Path(d)
        if not base_dir.is_dir():
            continue

        i = 1
        fails = []
        for f in base_dir.iterdir():
            ext = f.suffix.lower()
            if not f.is_file() or not ext in ext_list:
                continue

            timestamp = None
            if ext in (".jpg", ".jpeg"):
                try:
                    with Image.open(f) as img:
                        exif = {ExifTags.TAGS[k]: v for k, v in img._getexif().items() if k in ExifTags.TAGS}
                    if "DateTimeOriginal" in exif:
                        dto = exif["DateTimeOriginal"]

                        if re.match(r"\d{4}:\d{2}:\d{2} \d{2}:\d{2}:\d{2}", dto):
                            d, t = dto.split()
                            timestamp = d.replace(":", "-") + " " + t
                except Exception as e:
                    ColorLog.warn(f"Could not parse EXIF for {f}: {e}")

            desc = f"""\
            =={{{{int:filedesc}}}}==
            {{{{Information
            |description={base_dir.name}
            |date={timestamp or datetime.fromtimestamp(f.stat().st_mtime).strftime('%Y-%m-%d %H:%M:%S')}
            |source={{{{Own}}}}
            |author=~~~
            }}}}

            =={{{{int:license-header}}}}==
            {{{{Self|Cc-by-sa-4.0}}}}
            
            [[Category:{base_dir.name}]]
            [[Category:Files by {wiki.username}]]"""

            if not wiki.upload(str(f), f"{base_dir.name} {i} {date.today()}{ext}", dedent(desc), ""):
                fails.append(f)
            i += 1

    if fails:
        print("Failed to upload:")
        for f in fails:
            print(f)
    else:
        print("Complete, with no failures")


if __name__ == '__main__':
    main()



import base64
import re
import getpass

from pathlib import Path


class Wgen:

    @staticmethod
    def load_px(px_file=Path(".px.txt")):
        """Loads the specified password file if it exists.  Returns a dictionary with username/passwords that were found

        :param load_px: The path to the password file
        """
        pxf = Path.home() / px_file

        if not pxf.is_file():
            raise FileNotFoundError(f"{pxf} does not exist or is a directory. Did you run Wgen yet?  If there is a dir here, rename it first.")

        return dict([line.split("\t") for line in base64.b64decode(pxf.read_text().encode()).decode().strip().splitlines()])

    @staticmethod
    def setup(out_file=Path(".px.txt"), allow_continue=True):
        """Interactively creates a password file.

        :param out_file: The path to create the password file at.  CAVEAT: If a file exists at this location exists it will be overwritten.
        :param allow_continue: Set True to allow user to enter more than one user-pass combo.
        """
        pxl = {}

        while True:
            print("Please enter the username/password combo(s) you would like to use.")
            u = input("Username: ")
            p = getpass.getpass()
            confirm_p = getpass.getpass("Confirm Password: ")

            if p != confirm_p:
                print("ERROR: Entered passwords do not match")
                if not re.match("(?i)(y|yes)", input("Try again? (y/N): ")):
                    break
            else:
                pxl[u] = p

                if not allow_continue or not re.match("(?i)(y|yes)", input("Continue? (y/N): ")):
                    break

        if not pxl:
            print("WARNING: You did not make any entries.  Doing nothing.")
            return

        out_file.write_text(base64.b64encode("\n".join([f"{k}\t{v}" for k, v in pxl.items()]).encode()).decode())
        print(f"Successfully created '{out_file}'")



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="simple-commons-uploader",
    version="0.2.1",
    author="Fastily",
    author_email="fastily@users.noreply.github.com",
    description="Batch Wikimedia Commons Uploader",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/fastily/simple-commons-uploader",
    include_package_data=True,
    packages=setuptools.find_packages(),
    install_requires=['Pillow', 'requests'],
    entry_points={
        'console_scripts': [
            'scu = scu.__main__:main'
        ]
    },
    classifiers=[
        "Natural Language :: English",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
)